# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

from typing import Any, List, Tuple
from flax import struct

import jax
from jax.random import KeyArray
import jax.numpy as jnp

from .mdp import Timestep


class ReplayBuffer(struct.PyTreeNode):
    """A circular buffer used for Experience Replay (ER):
    Li, L., 1993, https://apps.dtic.mil/sti/pdfs/ADA261434.pdf.
    Use the `CircularBuffer.init` method to construct a buffer."""

    capacity: int = struct.field(pytree_node=False)
    """Returns the capacity of the buffer."""
    elements: Any = struct.field(pytree_node=True)
    """The elements currently stored in the buffer."""
    idx: jax.Array = struct.field(pytree_node=True)
    """The index of the next element to be added to the buffer."""

    @classmethod
    def create(cls, timestep: Timestep, capacity: int, n_steps: int=1) -> ReplayBuffer:
        """Constructs a CircularBuffer class."""
        # reserve memory
        uninitialised_elements = jax.tree_map(
            lambda x: jnp.broadcast_to(
                jnp.asarray(x * 0, dtype=x.dtype),
                (capacity, n_steps + 1, *jnp.asarray(x).shape),
            ),
            timestep,
        )
        return cls(
            capacity=capacity,
            elements=uninitialised_elements,
            idx=jnp.asarray(0),
        )

    def size(self) -> jax.Array:
        """Returns the number of elements currently stored in the buffer."""
        return self.idx

    def add(self, item: Any) -> ReplayBuffer:
        """Adds a single element to the buffer. If the buffer is full,
        the oldest element is overwritten."""
        idx = self.idx % self.capacity
        elements = jax.tree_map(lambda x, y: x.at[idx].set(y), self.elements, item)
        return self.replace(
            idx=self.idx + 1,
            elements=elements,
        )

    def sample(self, key: KeyArray, n: int = 1) -> Any:
        """Samples `n` elements uniformly at random from the buffer,
        and stacks them into a single pytree.
        If `n` is greater than state.idx,
        the function returns uninitialised elements"""
        indices = jax.random.randint(key=key, shape=(n,), minval=0, maxval=self.idx)
        items = jax.tree_map(lambda x: x[indices], self.elements)
        return items


class EpisodeBuffer(struct.PyTreeNode):
    """A asynchronous episodic memory buffer used for online learning."""

    elements: Any
    """The elements currently stored in the buffer."""
    idx: int = struct.field(pytree_node=False)
    """The index of the next element to be added to the buffer."""
    size: int = struct.field(pytree_node=False)

    @classmethod
    def create(cls, example_item: Any, size: int) -> EpisodeBuffer:
        """Constructs a CircularBuffer class."""
        # reserve memory
        uninitialised_elements = jax.tree_map(
            lambda x: jnp.broadcast_to(x * 0, (size, *x.shape)), example_item
        )
        return cls(elements=uninitialised_elements, idx=0, size=size)

    def add(self, item: Any) -> EpisodeBuffer:
        """Adds a single element to the buffer. If the buffer is full,
        the oldest element is overwritten."""
        # index updating requires jitting to guarantee in-place efficiency
        # see https://jax.readthedocs.io/en/latest/_autosummary/jax.numpy.ndarray.at.html
        elements = jax.tree_map(lambda x, y: x.at[self.idx].set(y), self.elements, item)
        return self.replace(elements=elements, idx=self.idx + 1)

    def add_range(self, items: List[Any]) -> EpisodeBuffer:
        """Adds more than one elements to the buffer. If the buffer is full,
        the oldest elements are overwritten."""
        start = self.idx % self.size
        end = start + len(items)
        # index updating requires jitting to guarantee in-place efficiency
        # see https://jax.readthedocs.io/en/latest/_autosummary/jax.numpy.ndarray.at.html
        elements = jax.tree_map(
            lambda x, y: x.at[start:end].set(jnp.stack(y)), self.elements, items
        )
        return self.replace(elements=elements, idx=self.idx + len(items))

    def sample(self, n: int = 1) -> Tuple[Any, EpisodeBuffer]:
        """Samples `n` elements uniformly at random from the buffer,
        and stacks them into a single pytree. If `n` is greater than state.idx,
        the function returns uninitialised elements"""
        indices = jnp.arange(self.idx)[:n]
        items = jax.tree_map(lambda x: x[indices], self.elements)
        return items, self



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Tuple

import jax
import jax.numpy as jnp
from jax import Array
from gymnax.environments.environment import Environment as GymnaxEnvironment, EnvParams
from gymnax.environments.environment import EnvParams
from gymnax.environments.spaces import Space as GymnaxSpace, gymnax_space_to_gym_space
from jax.random import KeyArray

from helx.base.spaces import Space, Continuous
from helx.base.mdp import Timestep, TRANSITION, TERMINATION, TRUNCATION

from .environment import EnvironmentWrapper
from .gym import to_helx as gym_to_helx


def to_helx(gym_space: GymnaxSpace) -> Space:
    gym_space = gymnax_space_to_gym_space(gym_space)  # type: ignore
    return gym_to_helx(gym_space)  # type: ignore


def timestep_from_gym(
    obs, state, reward, done, info, action: Array, t: Array
) -> Timestep:
    return Timestep(
        observation=jnp.asarray(obs),
        reward=jnp.asarray(reward),
        step_type=(TRANSITION, TERMINATION)[done],
        action=jnp.asarray(action),
        t=t,
        state=state,
        info=info,
    )


class GymnaxWrapper(EnvironmentWrapper):
    """Static class to convert between Gymnax environments and helx environments."""
    env: GymnaxEnvironment
    params: EnvParams

    @classmethod
    def wraps(cls, env: Tuple[GymnaxEnvironment, EnvParams]) -> GymnaxWrapper:
        env_, params = env
        return cls(
            env=env_,
            observation_space=to_helx(env_.observation_space(params)),
            action_space=to_helx(env_.action_space(params)),
            reward_space=Continuous(),
            params=params,
        )

    def reset(self, key: KeyArray) -> Timestep:
        obs, state = self.env.reset(key, self.params)
        return Timestep(
            t=jnp.asarray(0),
            observation=jnp.asarray(obs),
            reward=jnp.asarray(0.0),
            step_type=TRANSITION,
            action=jnp.asarray(-1),
            state=state,
        )

    def _step(self, key: KeyArray, timestep: Timestep, action: jax.Array) -> Timestep:
        obs, state, reward, done, info = self.env.step(
            key=key, state=timestep.state, action=action, params=self.params
        )
        idx = 2 * done + jnp.asarray(
            timestep.t > self.params.max_steps_in_episode, dtype=jnp.int32
        )  # out-of-bounds returns clamps to 2
        step_type = jax.lax.switch(
            idx,
            (
                lambda: TRANSITION,
                lambda: TRUNCATION,
                lambda: TERMINATION,
            ),
        )
        return Timestep(
            t=timestep.t + 1,
            observation=jnp.asarray(obs),
            reward=jnp.asarray(reward),
            step_type=step_type,
            action=jnp.asarray(action),
            state=state,
        )



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# pyright: reportPrivateImportUsage=false
from __future__ import annotations

from typing import Tuple

import distrax
import jax
import jax.numpy as jnp
import jax.tree_util as jtu
import optax
from flax import linen as nn
from flax import struct
from flax.core.scope import VariableDict as Params
from jax import Array
from jax.random import KeyArray
from optax import GradientTransformation
import rlax

from helx.base.modules import Lambda, Parallel, Split, Temperature
from helx.base.mdp import Timestep, TERMINATION
from helx.base.memory import ReplayBuffer
from helx.base.spaces import Continuous
from helx.base import losses

from .agent import Agent, AgentState, HParams, Log


class SACHParams(HParams):
    # network
    action_space: Continuous
    tau: float = 0.005
    # rl
    replay_start: int = 1000
    replay_memory_size: int = 1000
    update_frequency: int = 1
    target_network_update_frequency: int = 10000
    discount: float = 0.99
    n_steps: int = 1
    entropy_rewards: bool = False
    # sgd
    batch_size: int = 32
    learning_rate: float = 3e-4


class SACLog(Log):
    buffer_size: Array = jnp.asarray(0)
    critic_loss: Array = jnp.asarray(float("inf"))
    actor_loss: Array = jnp.asarray(float("inf"))
    temperature_loss: Array = jnp.asarray(float("inf"))
    total_loss: Array = jnp.asarray(float("inf"))
    policy_entropy: Array = jnp.asarray(float("inf"))
    alpha: Array = jnp.asarray(float("inf"))


class SACState(AgentState):
    params: Tuple[Params, Params, Params, Params] = struct.field(pytree_node=True)
    buffer: ReplayBuffer = struct.field(pytree_node=True)
    log: SACLog = struct.field(pytree_node=True)


class SAC(Agent):
    """Implements Soft Actor-Critic with gaussian policy for continuous action spaces.
    Haarnoja, Tuomas, et al. "Soft actor-critic: Off-policy maximum entropy deep
        reinforcement learning with a stochastic actor."
        International conference on machine learning. PMLR, 2018.
        https://arxiv.org/abs/1801.01290 for the base SAC.
    Haarnoja, Tuomas, et al. "Soft actor-critic algorithms and applications."
        arXiv preprint arXiv:1812.05905 (2018).
        https://arxiv.org/abs/1812.05905 for the tunable entropy temperature
    Haonanm Yu, et al. "Do you need reward entropy (in practice)?".
        arXiv preprint arXiv:2201.12434. 2022 Jan 28.
        https://arxiv.org/abs/2201.12434

    Args:
        network (AgentNetwork): the network to use for the agent.
            the `actor_net` is usually a `GaussianPolicy` for continuous action spaces.
            the `critic_net` is double Q network.
            the `extra_net` is usually a `Temperature` layer to automatically
            tune the entropy pull.
        optimiser (GradientTransformation): an optax optimiser to perform gradient descent.
        hparams (SACHparams): the hyperparameters for the agent.
        seed (int): the seed to use for the agent.
            Using the same seed will result in reproducible agent results.
    """

    hparams: SACHParams = struct.field(pytree_node=False)
    optimiser: GradientTransformation = struct.field(pytree_node=False)
    actor: nn.Module = struct.field(pytree_node=False)
    critic: nn.Module = struct.field(pytree_node=False)
    critic_target: nn.Module = struct.field(pytree_node=False)
    temperature: nn.Module = struct.field(pytree_node=False)

    @classmethod
    def create(
        cls,
        hparams: SACHParams,
        optimiser: optax.GradientTransformation,
        actor_backbone: nn.Module,
        critic_backbone: nn.Module,
    ) -> SAC:
        action_size = hparams.action_space.size
        reshape = Lambda(lambda x: x.reshape(hparams.action_space.shape))
        actor = nn.Sequential(
            [
                actor_backbone,
                Split(2),
                Parallel((nn.Dense(action_size), nn.Dense(action_size))),
                Parallel((reshape, jtu.tree_map(lambda x: x, reshape))),
            ]
        )
        critic = nn.Sequential([critic_backbone, nn.Dense(1)])
        critic_target = nn.Sequential([critic_backbone, nn.Dense(1)])
        temperature = Temperature()
        return SAC(
            hparams=hparams,
            optimiser=optimiser,
            actor=actor,
            critic=critic,
            critic_target=critic_target,
            temperature=temperature,
        )

    def init(self, timestep: Timestep, *, key: KeyArray) -> SACState:
        """Initialises the agent state.

        Args:
            key (KeyArray): a JAX PRNG key.

        Returns:
            SACState: the initial agent state.
        """
        key, k1, k2, k3 = jax.random.split(key, num=4)
        obs = self.hparams.obs_space.sample(key=key)
        params = (
            self.actor.init(k1, obs),
            self.critic.init(k2, obs),
            self.critic_target.init(k2, obs),
            self.temperature.init(k3),
        )
        opt_state = self.optimiser.init(params)
        buffer = ReplayBuffer.create(
            timestep, self.hparams.replay_memory_size, self.hparams.n_steps
        )
        return SACState(
            iteration=jnp.asarray(0),
            opt_state=opt_state,
            log=SACLog(),
            params=params,
            buffer=buffer,
        )

    def sample_action(
        self, train_state: SACState, obs: Array, *, key: KeyArray, eval: bool = False
    ) -> Array:
        params_actor = train_state.params[0]
        mu, logvar = jtu.tree_map(jnp.asarray, self.actor.apply(params_actor, obs))
        # SAC uses a tanh squashing function
        # see https://arxiv.org/pdf/1801.01290.pdf, Appendix C.
        # and https://github.com/openai/spinningup/blob/master/spinup/algos/pytorch/sac/core.py#L29
        policy = distrax.Transformed(
            distrax.Normal(mu, jnp.exp(logvar)), distrax.Tanh()
        )
        action = policy.sample(seed=key)
        return action

    def loss(
        self,
        params: Tuple[Params, Params, Params, Params],
        timesteps: Timestep,
        *,
        key: KeyArray,
    ) -> Tuple[Array, Tuple[Array, Array, Array, Array, Array]]:
        s_tm1 = timesteps.observation[:-1]
        s_t = timesteps.observation[1:]
        r_t = timesteps.reward[:-1][0]  # [0] because scalar
        terminal_tm1 = timesteps.step_type[:-1] != TERMINATION
        k1, k2 = jax.random.split(key, num=2)

        # params
        params_actor, params_critic, params_critic_target, params_temperature = params

        # autotuning temp
        temperature = jnp.asarray(self.temperature.apply(params_temperature))
        alpha = jax.lax.stop_gradient(temperature)

        # actor
        mu_tm1, logvar_tm1 = jtu.tree_map(
            jnp.asarray, self.actor.apply(params_actor, s_tm1)
        )
        policy_tm1 = distrax.Transformed(
            distrax.Normal(mu_tm1, jnp.exp(logvar_tm1)), distrax.Tanh()
        )
        a_tm1, logprobs_a_tm1 = policy_tm1.sample_and_log_prob(seed=k1)

        mu_t, logvar_t = jtu.tree_map(jnp.asarray, self.actor.apply(params_actor, s_t))
        policy_t = distrax.Transformed(
            distrax.Normal(mu_t, jnp.exp(logvar_t)), distrax.Tanh()
        )
        a_t, logprobs_a_t = policy_t.sample_and_log_prob(seed=k1)

        # critic
        q_tm1 = self.critic.apply(params_critic, s_tm1)
        qA_t, qB_t = jtu.tree_map(jnp.asarray, self.critic.apply(params_critic, s_t))
        q_target = jnp.min(jnp.asarray([qA_t, qB_t]), axis=0)
        critic_loss = losses.double_dqn_loss(
            timesteps,
            self.critic,
            params_critic,
            params_critic_target,
            self.hparams.discount,
        )

        # augment reward with policy entropy
        policy_entropy = -jnp.sum(probs_a_tm1 * logprobs_a_tm1, axis=-1)

        # critic loss
        qA_tm1, qB_tm1 = self.critic.apply(params_critic, s_tm1)
        qA_t, qB_t = self.critic.apply(params_critic, s_t)
        q_target = jnp.min(jnp.asarray([qA_t, qB_t]), axis=0)
        v_t = q_target - alpha * logprobs_a_t
        rlax.double_q_learning(
            q_tm1,
            a_tm1,
            r_t,
            discount_t,
            q_target,
        )

        qA_tm1, qB_tm1 = jnp.asarray(qA_tm1), jnp.asarray(qB_tm1)
        qA_t, qB_t = jnp.asarray(qA_t), jnp.asarray(qB_t)
        q_t = jnp.min(jnp.stack([qA_t, qB_t], axis=0), axis=0)
        v_t = q_t - alpha * logprobs_a_t
        q_target = jax.lax.stop_gradient(
            r_t + terminal_tm1 * self.hparams.discount * v_t
        )
        critic_loss = jnp.asarray(
            optax.l2_loss(qA_tm1, q_target) + optax.l2_loss(qB_tm1, q_target)
        )

        # actor loss
        q_tm1 = jax.lax.stop_gradient(
            jnp.min(jnp.stack([qA_tm1, qB_tm1], axis=0), axis=0)
        )
        rlax.policy_gradient_loss()
        actor_loss = jnp.asarray(alpha * logprobs_a_tm1 - q_tm1)

        # temperature loss
        target_entropy = -self.hparams.action_space.ndim
        logprobs_a_tm1 = jax.lax.stop_gradient(logprobs_a_tm1)
        temperature_loss = jnp.asarray(
            -(temperature * (logprobs_a_tm1 + target_entropy))
        )

        loss = actor_loss + critic_loss + temperature_loss
        aux = (actor_loss, critic_loss, temperature_loss, policy_entropy, alpha)
        return loss, aux

    def update(
        self,
        train_state: SACState,
        transition: Timestep,
        *,
        key: KeyArray,
    ) -> SACState:
        # update iteration
        iteration = jnp.asarray(train_state.iteration + 1, dtype=jnp.int32)

        # update memory
        buffer = train_state.buffer.add(transition)

        # update actor, critic and temperature
        def _sgd_step(params, opt_state):
            def _loss_fn(params, transition):
                loss, aux = jax.vmap(self.loss, in_axes=(None, 0))(params, transition)
                return jnp.mean(loss), aux

            transitions = buffer.sample(key, self.hparams.batch_size)
            (loss, aux), grads = jax.value_and_grad(_loss_fn, has_aux=True)(
                params, transitions
            )
            updates, opt_state = self.optimiser.update(grads, opt_state)
            params = optax.apply_updates(params, updates)
            return params, opt_state, loss, aux

        cond = buffer.size() < self.hparams.replay_memory_size
        cond = jnp.logical_or(cond, iteration % self.hparams.update_frequency != 0)
        params, opt_state, loss, aux = jax.lax.cond(
            cond,
            lambda p, o: _sgd_step(p, o),
            lambda p, o: (p, o, jnp.asarray(float("inf"))),
            train_state.params,
            train_state.opt_state,
        )

        # update log
        log = SACLog(
            iteration=iteration,
            actor_loss=aux[0],
            critic_loss=aux[1],
            temperature_loss=aux[2],
            total_loss=loss,
            policy_entropy=aux[3],
            alpha=aux[4],
            step_type=transition.step_type[-1],
            returns=train_state.log.returns
            + jnp.sum(
                self.hparams.discount ** transition.t[:-1] * transition.reward[:-1]
            ),
            buffer_size=buffer.size(),
        )

        # update train_state
        train_state = train_state.replace(
            iteration=iteration,
            opt_state=opt_state,
            params=params,
            buffer=buffer,
            log=log,
        )
        return train_state



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations
from typing import overload

import dm_env
import dm_env.specs
import jax
import jax.numpy as jnp
from jax.random import KeyArray

from helx.base.mdp import Timestep, TERMINATION, TRANSITION, TRUNCATION
from helx.base.spaces import Space, Discrete, Continuous
from .environment import EnvironmentWrapper


@overload
def to_helx(dm_space: dm_env.specs.DiscreteArray) -> Discrete:
    ...


@overload
def to_helx(dm_space: dm_env.specs.BoundedArray) -> Continuous:
    ...


@overload
def to_helx(dm_space: dm_env.specs.Array) -> Continuous:
    ...


def to_helx(dm_space: dm_env.specs.Array) -> Space:
    if isinstance(dm_space, dm_env.specs.DiscreteArray):
        return Discrete(dm_space.num_values)
    elif isinstance(dm_space, dm_env.specs.BoundedArray):
        return Continuous(
            shape=dm_space.shape,
            minimum=dm_space.minimum.min().item(),
            maximum=dm_space.maximum.max().item(),
        )
    elif isinstance(dm_space, dm_env.specs.Array):
        return Continuous(shape=dm_space.shape)
    else:
        raise NotImplementedError(
            "Cannot convert dm_env space of type {}".format(type(dm_space))
        )


def timestep_to_helx(
    timestep: dm_env.TimeStep, action: jax.Array, t: jax.Array
) -> Timestep:
    step_type = timestep.step_type
    obs = jnp.asarray(timestep.observation)
    reward = jnp.asarray(timestep.reward)
    discount = timestep.discount

    if timestep.step_type == dm_env.StepType.LAST:
        step_type = TERMINATION
    elif discount is not None and float(discount) == 0.0:
        step_type = TRUNCATION
    else:
        step_type = TRANSITION

    return Timestep(
        observation=obs,
        reward=reward,
        step_type=step_type,
        action=action,
        t=t,
        state=None,
    )


class DmEnvWrapper(EnvironmentWrapper):
    """Static class to convert between dm_env and helx environments."""
    env: dm_env.Environment

    @classmethod
    def wraps(cls, env: dm_env.Environment) -> DmEnvWrapper:
        self = cls(
            env=env,
            observation_space=to_helx(env.observation_spec()),  # type: ignore
            action_space=to_helx(env.action_spec()),  # type: ignore
            reward_space=to_helx(env.reward_spec()),  # type: ignore
        )
        return self

    def reset(self, key: KeyArray | int) -> Timestep:
        next_step = self.env.reset()
        return timestep_to_helx(next_step, jnp.asarray(-1), jnp.asarray(0))

    def _step(self, key: KeyArray, timestep: Timestep, action: jax.Array) -> Timestep:
        next_step = self.env.step(action.item())
        return timestep_to_helx(next_step, action, timestep.t + 1)



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""A set of functions to interoperate between the most common
RL environment interfaces, like `gym`, `gymnasium`, `dm_env`, `bsuite and others."""
from __future__ import annotations

import abc
from typing import Any

from flax import struct
import jax
from jax.random import KeyArray

from helx.base.mdp import Timestep, TRANSITION
from helx.base.spaces import Space


class Environment(struct.PyTreeNode):
    action_space: Space = struct.field(pytree_node=False)
    observation_space: Space = struct.field(pytree_node=False)
    reward_space: Space = struct.field(pytree_node=False)

    @abc.abstractmethod
    def reset(self, key: KeyArray) -> Timestep:
        raise NotImplementedError()

    @abc.abstractmethod
    def _step(self, key: KeyArray, timestep: Timestep, action: jax.Array) -> Timestep:
        raise NotImplementedError()

    def step(
        self, key: KeyArray | int, timestep: Timestep, action: jax.Array
    ) -> Timestep:
        # autoreset
        next_timestep = jax.lax.cond(
            timestep.step_type == TRANSITION,
            lambda timestep: self._step(key, timestep, action),
            lambda timestep: self.reset(key),
            timestep,
        )
        return next_timestep


class EnvironmentWrapper(Environment):
    env: Any = struct.field(pytree_node=False)

    @abc.abstractmethod
    def wraps(self, env: Any) -> Timestep:
        raise NotImplementedError()

    def unwrapped(self) -> Any:
        return self.env



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from flax.core.scope import VariableDict as Params
from jax import Array
import jax.numpy as jnp
import optax

from helx.base.mdp import Timestep, TERMINATION

from .dqn import DQNHParams, DQNLog, DQNState, DQN


class DDQNHParams(DQNHParams):
    ...


class DDQNLog(DQNLog):
    ...


class DDQNState(DQNState):
    ...


class DDQN(DQN):
    """Implements a Double Deep Q-Network:
    Deep Reinforcement Learning with Double Q-learning,
    Arthur Guez, David Silver, Hado van Hasselt
    https://ojs.aaai.org/index.php/AAAI/article/view/10295"""

    def loss(
        self,
        params: Params,
        timesteps: Timestep,
        params_target: Params,
    ) -> Array:
        s_tm1 = timesteps.observation[:-1]
        s_t = timesteps.observation[1:]
        a_tm1 = timesteps.action[:-1][0]  # [0] because scalar
        r_t = timesteps.reward[:-1][0]  # [0] because scalar
        terminal_tm1 = timesteps.step_type[:-1] != TERMINATION
        discount_t = self.hparams.discount ** timesteps.t[:-1][0]  # [0] because scalar
        q_tm1 = jnp.asarray(self.critic.apply(params, s_tm1))

        a_t = jnp.argmax(jnp.asarray(self.critic.apply(params, s_t)) * terminal_tm1)
        q_t = self.critic.apply(params_target, s_t)
        q_target = r_t + discount_t * q_t[a_t]

        td_loss = optax.l2_loss(q_tm1[a_tm1] - q_target)
        return jnp.asarray(td_loss)



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from functools import partial

import jax.numpy as jnp
import optax
from flax import struct
import flax.linen as nn

from .dqn import DQNHParams, DQNLog, DQNState, DQN
from helx.base.modules import Split, Merge, Parallel


class DuelingDQNHParams(DQNHParams):
    ...


class DuelingDQNLog(DQNLog):
    ...


class DuelingDQNState(DQNState):
    ...


class DuelingDQN(DQN):
    hparams: DuelingDQNHParams = struct.field(pytree_node=True)
    optimiser: optax.GradientTransformation = struct.field(pytree_node=True)
    critic: nn.Module = struct.field(pytree_node=True)

    @classmethod
    def create(
        cls,
        hparams: DuelingDQNHParams,
        optimiser: optax.GradientTransformation,
        backbone: nn.Module,
    ) -> DuelingDQN:
        critic = nn.Sequential(
            [
                backbone,
                Split(2),
                Parallel((nn.Dense(1), nn.Dense(hparams.action_space.maximum))),  # v, A
                Merge(partial(jnp.sum, axis=-1)) # q = v + A
            ]
        )
        return DuelingDQN(
            hparams=hparams,
            optimiser=optimiser,
            critic=critic,
        )



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from abc import abstractmethod

from jax import Array
from jax.random import KeyArray
import jax.numpy as jnp
from flax import struct
from optax import GradientTransformation, OptState

from helx.base.spaces import Space
from helx.base.mdp import Timestep, TRANSITION


class HParams(struct.PyTreeNode):
    obs_space: Space
    action_space: Space
    discount: float = 0.99
    n_steps: int = 1
    seed: int = 0


class Log(struct.PyTreeNode):
    iteration: Array = jnp.asarray(0)
    step_type: Array = TRANSITION
    returns: Array = jnp.asarray(0.0)


class AgentState(struct.PyTreeNode):
    iteration: Array
    opt_state: OptState
    log: Log


class Agent(struct.PyTreeNode):
    hparams: HParams = struct.field(pytree_node=False)
    optimiser: GradientTransformation = struct.field(pytree_node=False)

    @abstractmethod
    def create(*args) -> Agent:
        return

    @abstractmethod
    def init(self, key: KeyArray, timestep: Timestep) -> AgentState:
        raise NotImplementedError

    @abstractmethod
    def sample_action(
        self, agent_state: AgentState, obs: Array, *, key: KeyArray, eval: bool = False
    ) -> Array:
        raise NotImplementedError

    @abstractmethod
    def update(
        self,
        agent_state: AgentState,
        timesteps: Timestep,
        *,
        key: KeyArray,
    ) -> AgentState:
        raise NotImplementedError



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations
from typing import Sequence

import jax
import jax.numpy as jnp
from jax.random import KeyArray

from jax.core import Shape
from jax import Array
from jax.core import ShapedArray, Shape


MIN_INT = jax.numpy.iinfo(jnp.int16).min
MAX_INT = jax.numpy.iinfo(jnp.int16).max
MIN_INT_ARR = jnp.asarray(MIN_INT)
MAX_INT_ARR = jnp.asarray(MAX_INT)


class Space(ShapedArray):
    shape: Sequence[int]
    minimum: int = MIN_INT
    maximum: int = MAX_INT

    def __repr__(self):
        return "{}, min={}, max={})".format(
            super().__repr__()[:-1], self.minimum, self.maximum
        )

    def sample(self, key: KeyArray) -> Array:
        raise NotImplementedError()


class Discrete(Space):
    def __init__(self, n_elements: int = MAX_INT, shape: Shape = (), dtype=jnp.int32):
        super().__init__(shape, dtype)
        self.minimum = 0
        self.maximum = n_elements - 1

    def sample(self, key: KeyArray) -> Array:
        item = jax.random.randint(key, self.shape, self.minimum, self.maximum)
        # randint cannot draw jnp.uint, so we cast it later
        return jnp.asarray(item, dtype=self.dtype)


class Continuous(Space):
    def __init__(
        self, shape: Shape = (), minimum: int = MIN_INT, maximum: int = MAX_INT
    ):
        super().__init__(shape, jnp.float32)
        self.minimum = minimum
        self.maximum = maximum

    def sample(self, key: KeyArray) -> Array:
        assert jnp.issubdtype(self.dtype, jnp.floating)
        # see: https://github.com/google/jax/issues/14003
        lower = jnp.nan_to_num(self.minimum)
        upper = jnp.nan_to_num(self.maximum)
        return jax.random.uniform(
            key, self.shape, minval=lower, maxval=upper, dtype=self.dtype
        )



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import dataclasses
from os import PathLike
from typing import Sequence, get_type_hints

import chex
from absl import flags
from absl.flags import DEFINE, _argument_parser, _flagvalues

from .spaces import Space, Continuous, Discrete


class ShapeParser(_argument_parser.ListParser):
    def parse(self, shape: chex.Shape):
        return tuple(map(int, super().parse(shape)))


def DEFINE_shape(  # pylint: disable=invalid-name,redefined-builtin
    name, default, help, flag_values=_flagvalues.FLAGS, required=False, **args
):
    """Registers a flag whose value is a comma-separated list of dimensions.

    The flag value is parsed with a CSV parser.

    It is the same as a ListParser, but the output is a tuple of integers.

    Args:
      name: str, the flag name.
      default: list|str|None, the default value of the flag.
      help: str, the help message.
      flag_values: :class:`FlagValues`, the FlagValues instance with which the
        flag will be registered. This should almost never need to be overridden.
      required: bool, is this a required flag. This must be used as a keyword
        argument.
      **args: Dictionary with extra keyword args that are passed to the
        ``Flag.__init__``.

    Returns:
      a handle to defined flag.
    """
    parser = ShapeParser()
    serializer = _argument_parser.CsvListSerializer(",")
    return DEFINE(
        parser, name, default, help, flag_values, serializer, required=required, **args
    )


class SpaceParser(_argument_parser.ListParser):
    def parse(self, space: Space):
        shape = tuple(map(int, super().parse(space.shape)))
        if shape != (1,):
            space = Continuous(shape)
        else:
            space = Discrete(1)
        return (shape, space.dtype)


def DEFINE_space(  # pylint: disable=invalid-name,redefined-builtin
    name, default, help, flag_values=_flagvalues.FLAGS, required=False, **args
):
    """Registers a flag whose value is a comma-separated list of dimensions.

    The flag value is parsed with a CSV parser.

    It is the same as a ListParser, but the output is a tuple of integers.

    Args:
      name: str, the flag name.
      default: list|str|None, the default value of the flag.
      help: str, the help message.
      flag_values: :class:`FlagValues`, the FlagValues instance with which the
        flag will be registered. This should almost never need to be overridden.
      required: bool, is this a required flag. This must be used as a keyword
        argument.
      **args: Dictionary with extra keyword args that are passed to the
        ``Flag.__init__``.

    Returns:
      a handle to defined flag.
    """
    parser = SpaceParser()
    serializer = _argument_parser.CsvListSerializer(",")
    return DEFINE(
        parser, name, default, help, flag_values, serializer, required=required, **args
    )


def type_to_flag(type_):
    if type_ == bool:
        return flags.DEFINE_boolean
    elif type_ == int:
        return flags.DEFINE_integer
    elif type_ == float:
        return flags.DEFINE_float
    elif type_ == str:
        return flags.DEFINE_string
    elif type_ == chex.Shape:
        return DEFINE_shape
    elif type_ == Space:
        return DEFINE_space
    elif type_ in (list, Sequence, tuple):
        return flags.DEFINE_list
    else:
        raise ValueError("Type {} is not supported".format(type_))


def get_hparams_fields(type_):
    fields = get_type_hints(type_)
    if "obs_space" in fields:
        del fields["obs_space"]
    if "action_space" in fields:
        del fields["action_space"]
    return fields


def define_flags_from_hparams(type_):
    # get_type_hints to get precise types than annotations
    fields = get_hparams_fields(type_)
    if "obs_space" in fields:
        del fields["obs_space"]
    if "action_space" in fields:
        del fields["action_space"]

    # we wrap the type in a dataclass to get the default values
    if not hasattr(type_, "__dataclass_fields__"):
        raise TypeError(f"{type_} is not a dataclass")

    dataclass_fields = type_.__dataclass_fields__

    for field in fields:
        default_value = dataclass_fields[field].default

        if default_value == dataclasses.MISSING:
            default_value = None

        define_fn = type_to_flag(fields[field])
        define_fn(
            name=field,
            default=default_value,
            help=str(type.__doc__),
            required=False,
        )
    return


def hparams_from_flags(
    cls, obs_space: Space, action_space: Space, flags=flags.FLAGS, **kwargs
):
    hparams = {}
    for key, value in kwargs.items():
        hparams[key] = value
    for field in get_hparams_fields(cls):
        if field in kwargs:
            continue
        value = flags[field].value
        if value is None:
            raise ValueError("Flag {} is required".format(field))
        hparams[field] = value
    hparams["obs_space"] = obs_space
    hparams["action_space"] = action_space
    return cls(**hparams)


def hparams_from_yml(cls, obs_space: Space, action_space: Space, yml_path: PathLike):
    raise NotImplementedError()
    import yaml

    with open(yml_path, "r") as f:
        hparams = yaml.load(f, Loader=yaml.FullLoader)
    hparams["obs_space"] = obs_space
    hparams["action_space"] = action_space
    return cls(**hparams)



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations
from typing import Tuple, overload

import gymnasium
import gymnasium.spaces
from jax import Array
import jax.numpy as jnp
from jax.random import KeyArray
import numpy as np
from gymnasium.utils.step_api_compatibility import (
    TerminatedTruncatedStepType as GymnasiumTimestep,
)
from helx.base.mdp import Timestep, TRANSITION, TERMINATION, TRUNCATION
from helx.base.spaces import Continuous, Discrete, Space
from .environment import EnvironmentWrapper


@overload
def to_helx(gym_space: gymnasium.spaces.Discrete) -> Discrete:
    ...


@overload
def to_helx(gym_space: gymnasium.spaces.Box) -> Continuous:
    ...


def to_helx(gym_space: gymnasium.spaces.Space) -> Space:
    if isinstance(gym_space, gymnasium.spaces.Discrete):
        return Discrete(gym_space.n)
    elif isinstance(gym_space, gymnasium.spaces.Box):
        return Continuous(
            shape=gym_space.shape,
            minimum=gym_space.low.min().item(),
            maximum=gym_space.high.max().item(),
        )
    else:
        raise NotImplementedError(
            "Cannot convert dm_env space of type {}".format(type(gym_space))
        )


def timestep_from_gym(gym_step: GymnasiumTimestep, action: Array, t: Array) -> Timestep:
    obs, reward, terminated, truncated, _ = gym_step

    if terminated:
        step_type = TERMINATION
    elif truncated:
        step_type = TRUNCATION
    else:
        step_type = TRANSITION

    obs = jnp.asarray(obs)
    reward = jnp.asarray(reward)
    action = jnp.asarray(action)
    return Timestep(
        observation=obs,
        reward=reward,
        step_type=step_type,
        action=action,
        t=t,
        state=None,
    )


class GymnasiumWrapper(EnvironmentWrapper):
    """Static class to convert between gymnasium and helx environments."""
    env: gymnasium.Env

    @classmethod
    def wraps(cls, env: gymnasium.Env) -> GymnasiumWrapper:
        self = cls(
            env=env,
            observation_space=to_helx(env.observation_space),  # type: ignore
            action_space=to_helx(env.action_space),  # type: ignore
            reward_space=Continuous(
                (), minimum=int(env.reward_range[0]), maximum=int(env.reward_range[1])
            ),
        )
        return self

    def reset(self, seed: int | None = None) -> Timestep:
        timestep = self.env.reset(seed=seed)
        return timestep_from_gym(timestep, action=jnp.asarray(-1), t=jnp.asarray(0))

    def _step(self, key: KeyArray, timestep: Timestep, action: Array) -> Timestep:
        next_step = self.env.step(np.asarray(action))
        return timestep_from_gym(next_step, action, timestep.t + 1)



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

from typing import Tuple, overload

import gym
import gym.core
import gym.spaces
import gym.utils.seeding
from jax import Array
import jax.numpy as jnp
from jax.random import KeyArray
import numpy as np
from gym.utils.step_api_compatibility import TerminatedTruncatedStepType as GymTimestep

from helx.base.mdp import Timestep, TERMINATION, TRANSITION, TRUNCATION
from helx.base.spaces import Continuous, Discrete, Space
from .environment import EnvironmentWrapper


@overload
def to_helx(gym_space: gym.spaces.Discrete) -> Discrete:
    ...


@overload
def to_helx(gym_space: gym.spaces.Box) -> Continuous:
    ...


def to_helx(gym_space: gym.spaces.Space) -> Space:
    if isinstance(gym_space, gym.spaces.Discrete):
        return Discrete(gym_space.n)
    elif isinstance(gym_space, gym.spaces.Box):
        return Continuous(
            shape=gym_space.shape,
            minimum=gym_space.low.min().item(),
            maximum=gym_space.high.max().item(),
        )
    else:
        raise NotImplementedError(
            "Cannot convert dm_env space of type {}".format(type(gym_space))
        )


def timestep_from_gym(gym_step: GymTimestep, action: Array, t: Array) -> Timestep:
    obs, reward, terminated, truncated, _ = gym_step

    if terminated:
        step_type = TERMINATION
    elif truncated:
        step_type = TRUNCATION
    else:
        step_type = TRANSITION

    obs = jnp.asarray(obs)
    reward = jnp.asarray(reward)
    action = jnp.asarray(action)
    return Timestep(
        observation=obs,
        reward=reward,
        step_type=step_type,
        action=action,
        t=t,
        state=None,
    )


class GymWrapper(EnvironmentWrapper):
    """Static class to convert between gym and helx environments."""
    env: gym.Env

    @classmethod
    def wraps(cls, env: gym.Env) -> GymWrapper:
        self = cls(
            env=env,
            observation_space=to_helx(env.observation_space),  # type: ignore
            action_space=to_helx(env.action_space),  # type: ignore
            reward_space=to_helx(env.reward_range),  # type: ignore
        )
        return self

    def reset(self, seed: int | None = None) -> Timestep:
        try:
            timestep = self.env.reset(seed=seed)
        except TypeError:
            # TODO(epignatelli): remove try/except when gym3 is updated.
            # see: https://github.com/openai/gym3/issues/8
            timestep = self.env.reset()
        return timestep_from_gym(timestep, action=jnp.asarray(-1), t=jnp.asarray(0))

    def _step(self, key: KeyArray, timestep: Timestep, action: Array) -> Timestep:
        next_step = self.env.step(np.asarray(action))
        return timestep_from_gym(next_step, action, timestep.t + 1)



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from functools import partial

from typing import Callable, Sequence, Tuple

import flax.linen as nn
import jax.numpy as jnp
import jax.tree_util as jtu
from jax import Array


class Split(nn.Module):
    n_heads: int

    @nn.compact
    def __call__(self, x: Array) -> Tuple[Array, ...]:
        return (x,) * self.n_heads


class Merge(nn.Module):
    aggregate: Callable[[Array], Array] = partial(jnp.sum, axis=-1)

    @nn.compact
    def __call__(self, *arrays: Tuple[Array, ...]) -> Array:
        x = jnp.broadcast_arrays(*arrays)  # type: ignore
        x = jnp.asarray(x)
        return self.aggregate(x)


class Parallel(nn.Module):
    functions: Tuple[nn.Module, ...]

    @nn.compact
    def __call__(self, *inputs: Tuple[Array, ...]) -> Tuple[Array, ...]:
        return jtu.tree_map(lambda f, x: f(x), self.functions, inputs)


class Lambda(nn.Module):
    function: Callable[[Array], Array]

    @nn.compact
    def __call__(self, x: Array) -> Array:
        return self.function(x)


class Flatten(nn.Module):
    """A Flax module that flattens the input array."""

    @nn.compact
    def __call__(self, x: Array) -> Array:
        return x.flatten()


class Identity(nn.Module):
    """A Flax module that returns the input array."""

    @nn.compact
    def __call__(self, x: Array) -> Array:
        return x


class Temperature(nn.Module):
    initial_temperature: float = 1.0

    @nn.compact
    def __call__(self) -> Array:
        temperature = self.param(
            "temperature_value",
            init_fn=lambda key: jnp.log(self.initial_temperature),
        )
        return jnp.exp(temperature)


class MLP(nn.Module):
    """Defines a multi-layer perceptron by alternating a dense layer with a ReLU
    non-linearity.
    Args:
        features (Sequence[int]): A sequence of integers that define the number of
            units in each dense layer. E.g., (32, 64) defines a network with two
            dense layers, the first with 32 units and the second with 64.
    Example:
        >>> import jax.numpy as jnp
        >>> from helx.networks import MLP
        >>> mlp = MLP(features=(32, 64))
        >>> x = jnp.ones((1, 28, 28))
        >>> mlp(x).shape
        (1, 64)"""

    features: Sequence[int]
    activation: Callable[[Array], Array] = nn.relu

    @nn.compact
    def __call__(self, x: Array, *args, **kwargs) -> Array:
        for i in range(len(self.features)):
            x = nn.Dense(features=self.features[i])(x)
            x = self.activation(x)
        return x


class CNN(nn.Module):
    """Defines a convolutional neural network by alternating a
    convolution with a ReLU non-linearity.
    There is no pooling or other non-linearities between convolutions.
    Args:
        features (Tuple[int, ...]): A sequence of integers that define the number of
            filters in each convolution. E.g., (32, 64) defines a network with
            two convolutions, the first with 32 filters and the second with 64.
        kernel_sizes (Tuple[Tuple[int, ...]]): A sequence of tuples that define
            the kernel size of each convolution.
        strides (Tuple[Tuple[int, ...]]): A sequence of tuples that define the
            stride of each convolution.
        padding (Tuple[nn.linear.PaddingLike]): A sequence of padding modes for
            each convolution.

    Example:
        >>> import jax.numpy as jnp
        >>> from helx.networks import CNN
        >>> cnn = CNN(
        ...     features=(32, 64),
        ...     kernel_sizes=((3, 3), (3, 3)),
        ...     strides=((1, 1), (1, 1)),
        ...     padding=("SAME", "SAME"),
        ... )
        >>> cnn(jnp.ones((1, 10, 10, 3))).shape
        (1, 10, 10, 64)
    """

    features: Tuple[int, ...] = (32,)
    kernel_sizes: Tuple[Tuple[int, ...], ...] = ((3, 3),)
    strides: Tuple[Tuple[int, ...], ...] = ((1, 1),)
    paddings: Tuple[nn.linear.PaddingLike, ...] = ("SAME",)
    activation: Callable[[Array], Array] = nn.relu
    flatten: bool = False

    def setup(self):
        assert (
            len(self.features)
            == len(self.kernel_sizes)
            == len(self.strides)
            == len(self.paddings)
        ), "All arguments must have the same length, got {} {} {} {}".format(
            len(self.features),
            len(self.kernel_sizes),
            len(self.strides),
            len(self.paddings),
        )
        self.modules = [
            nn.Conv(
                features=self.features[i],
                kernel_size=self.kernel_sizes[i],
                strides=self.strides[i],
                padding=self.paddings[i],
            )
            for i in range(len(self.features))
        ]

    def __call__(self, x: Array, *args, **kwargs) -> Array:
        for module in self.modules:
            x = module(x)
            x = self.activation(x)
        if self.flatten:
            x = x.flatten()
        return x



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import jax
import jax.numpy as jnp
import numpy as np
from jax import Array
from jax.core import Shape


def imresize(
    img, size: Shape, method: str = "bilinear", channel_first: bool = True, **kwargs
):
    assert (
        len(size) == 2
    ), "Size must be a tuple of two numbers, respectively height and width, not {}".format(
        size
    )
    if channel_first:
        size = (*img.shape[:-2], *tuple(size))
    else:
        size = (*img.shape[:-3], *tuple(size), img.shape[-1])
    kwargs.update(shape=size, method=method)
    return jax.image.resize(img, **kwargs)


def greyscale(img: Array, channel_first: bool = False):
    """Perceptual luminance-preserving grayscale transformation"""
    if channel_first:
        shape = (3,) * (img.ndim - 1) + (1,)
    else:
        shape = (1,) * (img.ndim - 1) + (3,)
    luminance_mask = jnp.array([0.2126, 0.7152, 0.0722]).reshape(shape)
    img = jnp.sum(img * luminance_mask, axis=-1).squeeze()
    return img


def ensure_video_format(video: Array, channel_first: bool = True):
    """Ensure that the video has the right shape for `wandb.Video`
    Assumes that the first axis is the time axis.
    Args:
        video (Array): the video to be reshaped.
        channel_first (bool): whether the input array has channel first or not.
    Returns:
        (Array): the reshaped video with shape (time, channel, height, width),
        or None if `video.ndim < 3` or `video.ndim > 5`."""
    # Check shape first
    n_dim = video.ndim
    if n_dim < 3 or n_dim > 5:
        logging.warning(
            "Video must have at least three and at most four dimensions, got {} instead.".format(
                n_dim
            )
        )
        return None

    # video.ndim either 3 or 4
    if n_dim == 3:
        # add channel axis
        video = jnp.expand_dims(video, axis=1)
        video = jnp.repeat(video, repeats=3, axis=1)
    # if 4
    elif not channel_first:
        video = video.transpose(0, 3, 1, 2)

    # Check dtype and range
    if video.dtype == jnp.float32:
        video = jnp.clip(video, 0, 1)
        video = (video * 255).astype(np.uint8)
    return np.array(video)



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

from typing import Any, Dict

from jax import Array
import jax.numpy as jnp
from flax import struct


TRANSITION = jnp.asarray(0)
TRUNCATION = jnp.asarray(1)
TERMINATION = jnp.asarray(2)


class Timestep(struct.PyTreeNode):
    t: Array
    """The number of timesteps elapsed from the last reset of the environment"""
    observation: Array
    """The observation corresponding to the current state (for POMDPs)"""
    action: Array
    """The action taken by the agent at the current timestep a_t = $\\pi(s_t)$, where $s_t$ is `state`"""
    reward: Array
    """The reward $r_{t=1}$ received by the agent after taking action $a_t$"""
    step_type: Array
    """The type of the current timestep (see `helx.mdp`)"""
    state: Any
    """The true state of the MDP, $s_t$ before taking action `action`"""
    info: Dict[str, Any] = struct.field(default_factory=dict)
    """Additional information about the environment. Useful for accumulations (e.g. returns)"""



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import jax
import jax.numpy as jnp
import brax.envs
from jax.random import KeyArray

from helx.base.spaces import MAX_INT_ARR, Continuous
from helx.base.mdp import Timestep, TRANSITION, TERMINATION
from .environment import EnvironmentWrapper


class BraxWrapper(EnvironmentWrapper):
    """Static class to convert between Brax environments and helx environments.
    The current environments are supported:
    ['ant', 'halfcheetah', 'hopper', 'humanoid', 'humanoidstandup',
    'inverted_pendulum', 'inverted_double_pendulum', 'pusher',
    'reacher', 'walker2d']"""
    env: brax.envs.Env

    @classmethod
    def wraps(cls, env: brax.envs.Env) -> BraxWrapper:
        return cls(
            env=env,
            observation_space=Continuous(shape=(env.observation_size,)),
            action_space=Continuous(shape=(env.action_size,)),
            reward_space=Continuous(),
        )

    def reset(self, key: KeyArray) -> Timestep:
        state = self.env.reset(key)
        return Timestep(
            t=jnp.asarray(0),
            observation=state.obs,
            reward=state.reward,
            step_type=TRANSITION,
            action=self.action_space.sample(key),
            state=state.pipeline_state,
            info={**state.info, **state.metrics}
        )

    def _step(self, key: KeyArray, timestep: Timestep, action: jax.Array) -> Timestep:
        # unwrap
        state = brax.envs.State(
            pipeline_state=timestep.state,
            obs=timestep.observation,
            reward=timestep.reward,
            done=timestep.step_type == TERMINATION,
            info=timestep.info,
            metrics=timestep.info
        )

        # step
        state = self.env.step(state=state, action=action)

        # wrap again
        truncation = state.info.get("truncation", state.metrics.get("truncation", MAX_INT_ARR))
        step_type = 2 * state.done + timestep.t > truncation
        return Timestep(
            t=timestep.t + 1,
            observation=state.obs,
            reward=state.reward,
            step_type=jnp.asarray(step_type, dtype=jnp.int32),
            action=action,
            state=state.pipeline_state,
            info={**state.info, **state.metrics}
        )



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import distrax
import jax
import jax.numpy as jnp
import jax.tree_util as jtu
import optax
import rlax
from flax import linen as nn
from flax import struct
from flax.core.scope import VariableDict as Params
from jax import Array
from jax.random import KeyArray
from optax import GradientTransformation

from helx.base.mdp import TERMINATION, Timestep
from helx.base.memory import ReplayBuffer
from helx.base.spaces import Discrete
from helx.base import losses
from .agent import Agent, HParams, Log, AgentState


class DQNHParams(HParams):
    # network
    hidden_size: int = 128
    # rl
    initial_exploration: float = 1.0
    final_exploration: float = 0.01
    final_exploration_frame: int = 1000000
    replay_start: int = 1000
    replay_memory_size: int = 1000
    update_frequency: int = 1
    target_network_update_frequency: int = 10000
    discount: float = 0.99
    n_steps: int = 1
    # sgd
    batch_size: int = 32
    learning_rate: float = 0.00025
    gradient_momentum: float = 0.95
    squared_gradient_momentum: float = 0.95
    min_squared_gradient: float = 0.01


class DQNLog(Log):
    buffer_size: Array = jnp.asarray(0)
    critic_loss: Array = jnp.asarray(float("inf"))


class DQNState(AgentState):
    params: Params = struct.field(pytree_node=True)
    params_target: Params = struct.field(pytree_node=True)
    buffer: ReplayBuffer = struct.field(pytree_node=True)
    log: DQNLog = struct.field(pytree_node=True)


class DQN(Agent):
    """Implements a Deep Q-Network:
    Mnih, Volodymyr, et al. "Human-level control through deep reinforcement learning."
    Nature 518.7540 (2015): 529-533.
    https://www.nature.com/articles/nature14236"""

    hparams: DQNHParams = struct.field(pytree_node=False)
    optimiser: GradientTransformation = struct.field(pytree_node=False)
    critic: nn.Module = struct.field(pytree_node=False)

    @classmethod
    def create(
        cls,
        hparams: DQNHParams,
        optimiser: optax.GradientTransformation,
        backbone: nn.Module,
    ) -> DQN:
        critic = nn.Sequential(
            [
                backbone,
                nn.Dense(hparams.action_space.maximum),
            ]
        )
        return DQN(
            hparams=hparams,
            optimiser=optimiser,
            critic=critic,
        )

    def init(self, timestep: Timestep, *, key: KeyArray) -> DQNState:
        hparams = self.hparams
        assert isinstance(hparams.action_space, Discrete)
        iteration = jnp.asarray(0)
        params = self.critic.init(
            key, hparams.obs_space.sample(key)
        )
        params_target = jtu.tree_map(lambda x: x, params)  # copy params
        buffer = ReplayBuffer.create(
            timestep,
            hparams.replay_memory_size,
            hparams.n_steps,
        )
        opt_state = self.optimiser.init(params)
        return DQNState(
            iteration=iteration,
            params=params,
            params_target=params_target,
            opt_state=opt_state,
            buffer=buffer,
            log=DQNLog(),
        )

    def sample_action(
        self, train_state: DQNState, obs: Array, *, key: KeyArray, eval: bool = False
    ) -> Array:
        q_values = self.critic.apply(train_state.params, obs)
        eps = optax.polynomial_schedule(
            init_value=self.hparams.initial_exploration,
            end_value=self.hparams.final_exploration,
            transition_steps=self.hparams.final_exploration_frame
            - self.hparams.replay_start,
            transition_begin=self.hparams.replay_start,
            power=1.0,
        )(train_state.iteration)
        action = distrax.EpsilonGreedy(jnp.asarray(q_values), jnp.asarray(eps)).sample(seed=key)  # type: ignore
        return action

    def loss(
        self,
        params: Params,
        timesteps: Timestep,
        params_target: Params,
    ) -> Array:
        s_tm1 = timesteps.observation[:-1]
        s_t = timesteps.observation[1:]
        a_tm1 = timesteps.action[:-1][0]  # [0] because scalar
        r_t = timesteps.reward[:-1][0]  # [0] because scalar
        terminal_tm1 = timesteps.step_type[:-1] != TERMINATION
        discount_t = self.hparams.discount ** timesteps.t[:-1][0]  # [0] because scalar

        q_tm1 = jnp.asarray(self.critic.apply(params, s_tm1))
        q_t = jnp.asarray(self.critic.apply(params_target, s_t)) * terminal_tm1

        td_error = rlax.q_learning(
            q_tm1, a_tm1, r_t, discount_t, q_t, stop_target_gradients=True
        )
        td_error = jnp.asarray(td_error)
        td_loss = jnp.mean(0.5 * td_error**2)
        return td_loss

    def update(
        self,
        train_state: DQNState,
        transition: Timestep,
        *,
        key: KeyArray,
    ) -> DQNState:
        # update iteration
        iteration = jnp.asarray(train_state.iteration + 1, dtype=jnp.int32)

        # update memory
        buffer = train_state.buffer.add(transition)

        # update critic
        def _sgd_step(params, opt_state):
            transitions = buffer.sample(key, self.hparams.batch_size)
            loss_fn = lambda params, trans: jnp.mean(
                jax.vmap(self.loss, in_axes=(None, 0, None))(
                    params, trans, train_state.params_target
                )
            )
            loss, grads = jax.value_and_grad(loss_fn)(params, transitions)
            updates, opt_state = self.optimiser.update(grads, opt_state)
            params = optax.apply_updates(params, updates)
            return params, opt_state, loss

        cond = buffer.size() < self.hparams.replay_memory_size
        cond = jnp.logical_or(cond, iteration % self.hparams.update_frequency != 0)
        params, opt_state, loss = jax.lax.cond(
            cond,
            lambda p, o: _sgd_step(p, o),
            lambda p, o: (p, o, jnp.asarray(float("inf"))),
            train_state.params,
            train_state.opt_state,
        )

        # update target critic
        params_target = optax.periodic_update(
            params,
            train_state.params_target,
            jnp.asarray(iteration, dtype=jnp.int32),
            self.hparams.target_network_update_frequency,
        )

        # log
        log = DQNLog(
            iteration=iteration,
            critic_loss=loss,
            step_type=transition.step_type[-1],
            returns=train_state.log.returns
            + jnp.sum(
                self.hparams.discount ** transition.t[:-1] * transition.reward[:-1]
            ),
            buffer_size=buffer.size(),
        )

        # update train_state
        train_state = train_state.replace(
            iteration=iteration,
            opt_state=opt_state,
            params=params,
            params_target=params_target,
            buffer=buffer,
            log=log,
        )
        return train_state



import logging
from pprint import pformat
from typing import TypeVar
import logging

import wandb

import jax.numpy as jnp
import jax.experimental.host_callback as hcb
import flax.linen as nn

from helx.agents import Log
from helx.base.mdp import TRANSITION


T = TypeVar("T", bound=nn.Module)

class Color:
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    DARKCYAN = '\033[36m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    RESET = '\033[0m'
    WHITE = '\033[1;37m'


def log_wandb(logs: Log) -> Log:
    log_dict = logs.__dict__
    for k, v in log_dict.items():
        if k.startswith("_"):
            continue
        if jnp.isnan(v):
            logging.warning(f"NaN value for {k}")
        # do not log returns unless episode has finished
        if k == "returns":
            if not "step_type" in log_dict:
                raise ValueError("Log must have step_type to log returns")
            if log_dict["step_type"] == TRANSITION:
                continue
        wandb.log({k: v}, commit=False)
    wandb.log({})  # commit flush
    return logs


def host_log_wandb(logs: Log) -> Log:
    hcb.id_tap(lambda x, _: log_wandb(x), logs)
    return logs


def log_start(seed, agent, env, budget):
    logging.info(Color.BOLD + Color.WHITE + "Experiment starts with seed {} and budget {}".format(seed, budget) + Color.RESET)
    logging.info(pformat(agent))
    logging.info(pformat(env))


def log_end(seed, agent, env, budget):
    logging.info(Color.BOLD + Color.WHITE + "Experiment completed with seed {} and budget {}".format(seed, budget) + Color.RESET)
    logging.info(pformat(agent))
    logging.info(pformat(env))


def report(run_fn):
    def wrapped(seed, agent, env, budget):
        log_start(seed, agent, env, budget)
        results = run_fn(seed, agent, env, budget)
        log_end(seed, agent, env, budget)
        return results
    return wrapped



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import flax.linen as nn
import navix as nx
import optax
from absl import app, flags

import helx
import wandb

helx.base.config.define_flags_from_hparams(helx.agents.SACDHParams)
FLAGS = flags.FLAGS


def main(argv):
    wandb.init(mode="disabled")

    # environment
    env = nx.environments.KeyDoor(5, 10, 10, observation_fn=nx.observations.categorical)
    env = helx.envs.interop.to_helx(env)

    # optimiser
    optimiser = optax.rmsprop(
        learning_rate=FLAGS.learning_rate,
        momentum=FLAGS.gradient_momentum,
        eps=FLAGS.min_squared_gradient,
        centered=True,
    )

    # agent
    hparams = helx.base.config.hparams_from_flags(
        helx.agents.SACDHParams,
        obs_space=env.observation_space,
        action_space=env.action_space,
        replay_start=10,
        batch_size=2,
    )
    actor_backbone = nn.Sequential(
        [
            helx.base.modules.Flatten(),
            helx.base.modules.MLP(features=[32, 16]),
        ]
    )
    critic_backbone = nn.Sequential(
        [
            helx.base.modules.Flatten(),
            helx.base.modules.MLP(features=[32, 16]),
        ]
    )
    agent = helx.agents.SACD.create(
        hparams=hparams,
        optimiser=optimiser,
        actor_backbone=actor_backbone,
        critic_backbone=critic_backbone,
    )

    helx.experiment.jrun(seed=FLAGS.seed, agent=agent, env=env, budget=20)


if __name__ == "__main__":
    app.run(main)



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import jax
import jax.numpy as jnp
from chex import Array

from .agent import Agent, HParams
from helx.base.mdp import Timestep


class Random(Agent):
    """Implements a Deep Q-Network:
    Mnih, Volodymyr, et al. "Human-level control through deep reinforcement learning."
    Nature 518.7540 (2015): 529-533.
    https://www.nature.com/articles/nature14236"""

    def __init__(
        self,
        hparams: HParams,
        seed: int,
    ):
        self.key: jax.random.KeyArray = jax.random.PRNGKey(seed)
        self.hparams: HParams = hparams
        self.iteration: int = 0

    def sample_action(self, observation: Array, eval: bool = False, **kwargs) -> Array:
        return self.hparams.action_space.sample(self.key)

    def update(self, episode: Timestep) -> Array:
        return jnp.asarray(())



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any, Tuple, overload

import bsuite.environments
import dm_env
import gym.core
import gym3.interop
import gymnasium.core
from gymnax.environments.environment import Environment as GymnaxEnvironment, EnvParams
import brax.envs
import navix as nx

from .environment import EnvironmentWrapper
from .bsuite import BsuiteWrapper
from .dm_env import DmEnvWrapper
from .gym import GymWrapper
from .gymnasium import GymnasiumWrapper
from .gymnax import GymnaxWrapper
from .brax import BraxWrapper
from .navix import NavixWrapper


@overload
def to_helx(env: gymnasium.core.Env) -> GymnasiumWrapper:
    ...


@overload
def to_helx(env: gym.core.Env) -> GymWrapper:
    ...


@overload
def to_helx(env: gym3.interop.ToGymEnv) -> GymWrapper:
    ...


@overload
def to_helx(env: dm_env.Environment) -> DmEnvWrapper:
    ...


@overload
def to_helx(env: bsuite.environments.Environment) -> BsuiteWrapper:
    ...


@overload
def to_helx(env: Tuple[GymnaxEnvironment, EnvParams]) -> GymnaxWrapper:
    ...


@overload
def to_helx(env: brax.envs.Env) -> BraxWrapper:
    ...


@overload
def to_helx(env: nx.environments.Environment) -> NavixWrapper:
    ...


def to_helx(env: Any) -> EnvironmentWrapper:
    # getting root env type for interop
    env_for_type = env
    while hasattr(env_for_type, "unwrapped") and env_for_type.unwrapped != env_for_type:
        env_for_type = env_for_type.unwrapped

    # converting the actual env, rather than the root env
    # which would remove time limits and o
    if isinstance(env_for_type, gymnasium.core.Env):
        return GymnasiumWrapper.wraps(env)
    elif isinstance(env_for_type, gym.core.Env):
        return GymWrapper.wraps(env)
    elif isinstance(env_for_type, gym3.interop.ToGymEnv):
        return GymWrapper.wraps(env)
    elif isinstance(env_for_type, dm_env.Environment):
        return DmEnvWrapper.wraps(env)
    elif isinstance(env_for_type, bsuite.environments.Environment):
        return BsuiteWrapper.wraps(env)
    elif (
        isinstance(env_for_type, tuple)
        and issubclass(type(env[0]), GymnaxEnvironment)
        # and issubclass(type(env[1]), EnvParams)  # gymnax EnvParams do not have a base class
    ):
        return GymnaxWrapper.wraps(env)
    elif isinstance(env_for_type, brax.envs.Env):
        return BraxWrapper.wraps(env)
    elif isinstance(env_for_type, nx.environments.Environment):
        return NavixWrapper.wraps(env)
    else:
        raise TypeError(
            f"Environment type {type(env)} is not supported. "
            "Only gymnasium, gym, dm_env and bsuite environments are supported."
        )



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import overload

import jax
import jax.numpy as jnp
import navix as nx
from jax.random import KeyArray

from helx.base.spaces import Continuous, Discrete, Space
from helx.base.mdp import TRANSITION, Timestep
from .environment import EnvironmentWrapper


@overload
def to_helx(space: nx.spaces.Discrete) -> Discrete:
    ...


@overload
def to_helx(space: nx.spaces.Continuous) -> Continuous:
    ...


@overload
def to_helx(space: nx.spaces.Space) -> Space:
    ...


def to_helx(space: nx.spaces.Space) -> Space:
    if isinstance(space, nx.spaces.Discrete):
        return Discrete(space.maximum.item(), shape=space.shape, dtype=space.dtype)
    elif isinstance(space, nx.spaces.Continuous):
        return Continuous(
            shape=space.shape,
            minimum=space.minimum.item(),
            maximum=space.maximum.item(),
        )
    else:
        raise NotImplementedError(
            "Cannot convert dm_env space of type {}".format(type(space))
        )


class NavixWrapper(EnvironmentWrapper):
    """Static class to convert between Gymnax environments and helx environments."""
    env: nx.environments.Environment

    @classmethod
    def wraps(cls, env: nx.environments.Environment) -> NavixWrapper:
        return cls(
            env=env,
            observation_space=to_helx(env.observation_space),
            action_space=to_helx(env.action_space),
            reward_space=Continuous(),
        )

    def reset(self, key: KeyArray) -> Timestep:
        timestep = self.env.reset(key)
        return Timestep(
            t=jnp.asarray(0),
            observation=timestep.observation,
            reward=timestep.reward,
            step_type=TRANSITION,
            action=jnp.asarray(-1),
            state=timestep.state,
            info=timestep.info
        )

    def _step(self, key: KeyArray, timestep: Timestep, action: jax.Array) -> Timestep:
        current_step = nx.environments.environment.Timestep(
            t=timestep.t,
            observation=timestep.observation,
            reward=timestep.reward,
            step_type=timestep.step_type,
            action=action,
            state=timestep.state,
            info=timestep.info
        )
        nexst_step = self.env.step(current_step, action)
        return Timestep(
            t=nexst_step.t,
            observation=nexst_step.observation,
            reward=nexst_step.reward,
            step_type=nexst_step.step_type,
            action=action,
            state=nexst_step.state,
            info=nexst_step.info
        )




# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Tuple
import jax
from jax.random import KeyArray
import jax.numpy as jnp
import wandb

from helx.base.mdp import Timestep, TRANSITION
from helx.agents import Agent, AgentState
from helx.envs.environment import Environment

from .logging import host_log_wandb, report


def run_episode(
    agent: Agent,
    agent_state: AgentState,
    env: Environment,
    eval: bool = False,
    *,
    key: KeyArray,
) -> Timestep:
    key, k1 = jax.random.split(key)
    timestep = env.reset(k1)
    timesteps = [timestep]
    while timestep.step_type == TRANSITION:
        key, k1, k2 = jax.random.split(key, 3)
        action = agent.sample_action(
            agent_state, timestep.observation, key=k1, eval=eval
        )
        timestep = env.step(k2, timestep, action)
        timesteps.append(timestep)

    # convert list of timesteps into a batched timestep object
    timesteps = jax.tree_util.tree_map(lambda *x: jnp.stack(x), *timesteps)
    return timesteps


def run_n_steps(
    agent: Agent,
    env: Environment,
    agent_state: AgentState,
    env_state: Timestep,
    n_steps: int,
    *,
    key: KeyArray,
    eval: bool = False,
) -> Timestep:
    timesteps = [env_state]
    for _ in range(n_steps):
        key, k1, k2 = jax.random.split(key, num=3)
        action = agent.sample_action(
            agent_state, env_state.observation, key=k1, eval=eval
        )
        env_state = env.step(k2, env_state, action)
        timesteps.append(env_state)

    # convert list of timesteps into a batched timestep object
    timesteps = jax.tree_util.tree_map(lambda *x: jnp.stack(x), *timesteps)
    return timesteps


@report
def run(
    seed: jax.Array,
    agent: Agent,
    env: Environment,
    budget: int,
) -> Tuple[AgentState, Timestep]:
    # init
    key = jax.random.PRNGKey(seed)
    key, k1, k2 = jax.random.split(key, num=3)
    env_state = env.reset(key=k1)
    agent_state = agent.init(key=k2, timestep=env_state)
    wandb.init(mode="disabled")

    for _ in range(budget):
        key, k1, k2 = jax.random.split(key, num=3)
        timesteps = run_n_steps(
            agent, env, agent_state, env_state, n_steps=agent.hparams.n_steps, key=key
        )
        agent_state = agent.update(agent_state, timesteps, key=key)

        host_log_wandb(
            agent_state.log
        )  # potentially blocking, this call is on the host, not on the device, despite jitting

    return agent_state, env_state


@report
def jrun(
    seed: jax.Array,
    agent: Agent,
    env: Environment,
    budget: int,
) -> Tuple[AgentState, Timestep]:
    def body_fun(
        val: Tuple[AgentState, Timestep, KeyArray]
    ) -> Tuple[AgentState, Timestep, KeyArray]:
        agent_state, env_state, key = val
        key, k1, k2 = jax.random.split(key, num=3)
        timesteps = run_n_steps(
            agent, env, agent_state, env_state, n_steps=agent.hparams.n_steps, key=k1
        )
        agent_state = agent.update(agent_state, timesteps, key=k2)
        host_log_wandb(
            agent_state.log
        )  # potentially blocking, this call is on the host, not on the device, despite jitting
        return agent_state, env_state, key

    # init
    key = jax.random.PRNGKey(seed)
    key, k1, k2 = jax.random.split(key, num=3)
    env_state = env.reset(key=k1)
    agent_state = agent.init(key=k2, timestep=env_state)
    wandb.init(mode="disabled")

    agent_state, env_state, _ = jax.lax.while_loop(
        lambda x: x[0].iteration < budget,
        body_fun,
        (
            agent_state,
            env_state,
            key,
        ),
    )
    return agent_state, env_state



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from typing import Tuple

import flax.linen as nn
import jax
import jax.numpy as jnp
import jax.tree_util as jtu
import optax
from flax.core.scope import VariableDict as Params
from jax import Array
from jax.random import KeyArray

from helx.base.modules import Temperature, Split, Parallel
from helx.base.mdp import Timestep, TERMINATION
from helx.base.spaces import Discrete

from .sac import SAC, SACHParams, SACState


class SACDHParams(SACHParams):
    action_space: Discrete


class SACD(SAC):
    """Soft Actor-Critic agent with softmax policy for discrete action spaces
    and a twinned Q-network for value approximation with a separate target network
    updated by polyak averaging the online network parameters.
    See:
        1. Haarnoja T., 2018 (https://arxiv.org/abs/1801.01290)
            for the base SAC
        2. Haarnoja T., 2019 (https://arxiv.org/abs/1812.05905)
            for the tunable entropy temperature and the ablation of the value network
        3. Christodoulou P., 2019 (https://arxiv.org/abs/1910.07207)
            for the adaptation to discrete action spaces
    """

    @classmethod
    def create(
        cls,
        hparams: SACDHParams,
        optimiser: optax.GradientTransformation,
        actor_backbone: nn.Module,
        critic_backbone: nn.Module,
    ) -> SACD:
        n_actions = hparams.action_space.maximum
        actor = nn.Sequential([actor_backbone, nn.Dense(n_actions)])
        critic = nn.Sequential(
            [
                Split(2),
                Parallel((critic_backbone, jtu.tree_map(lambda x: x, critic_backbone))),
                Parallel((nn.Dense(n_actions), nn.Dense(n_actions))),
            ]
        )
        temperature = Temperature()
        return SACD(
            hparams=hparams,
            optimiser=optimiser,
            actor=actor,
            critic=critic,
            temperature=temperature,
        )

    def sample_action(
        self, train_state: SACState, obs: Array, *, key: KeyArray, eval: bool = False
    ) -> Array:
        params_actor, _, _ = train_state.params
        logits = jnp.asarray(self.actor.apply(params_actor, obs))
        action = jax.random.categorical(key, logits)
        return action

    def loss(
        self,
        params: Tuple[Params, Params, Params],
        transition: Timestep,
    ) -> Tuple[Array, Tuple[Array, Array, Array, Array, Array]]:
        s_tm1 = transition.observation[:-1]
        s_t = transition.observation[1:]
        r_t = transition.reward[:-1][0]  # [0] because scalar
        terminal_tm1 = transition.step_type[:-1] != TERMINATION

        # params
        batch_matmul = lambda a, b: jnp.sum(a * b, axis=-1, keepdims=True)
        params_actor, params_critic, params_temperature = params

        # current estimates
        temperature = self.temperature(params_temperature)
        alpha = jax.lax.stop_gradient(temperature)
        logits_a_tm1 = self.actor(params_actor, s_tm1)
        logits_a_t = self.actor(params_actor, s_t)
        logprobs_a_tm1 = jax.nn.log_softmax(logits_a_tm1)
        logprobs_a_t = jax.nn.log_softmax(logits_a_t)
        probs_a_tm1 = jnp.exp(logprobs_a_tm1)

        # augment reward with policy entropy
        policy_entropy = -jnp.sum(probs_a_tm1 * logprobs_a_tm1, axis=-1)
        # SACLite: https://arxiv.org/abs/2201.12434
        policy_entropy = policy_entropy * self.hparams.entropy_rewards
        r_t = r_t + alpha * policy_entropy

        # critic is a double Q network
        qA_tm1, qB_tm1 = self.critic(params_critic, s_tm1)
        qA_t, qB_t = self.critic(params_critic, s_t)

        # actor_target: $V(s_t) = \pi(s_t)^T \cdot [\alpha \log(\pi(s_t)) + Q(s_t)]$
        q_tm1 = jax.lax.stop_gradient(
            jnp.min(jnp.stack([qA_tm1, qB_tm1], axis=0), axis=0)
        )
        actor_loss = batch_matmul(probs_a_tm1, alpha * logprobs_a_tm1 - q_tm1)

        # critic target: $V(s_{t+1}) = \pi(s_{t+1})^T \cdot [Q(s_{t+1}) - \alpha \log(\pi(s_{t+1}))]$
        probs_a_t = jnp.exp(logprobs_a_t)
        q_t = jnp.min(jnp.stack([qA_t, qB_t], axis=0), axis=0)
        # SAC calculates the expectation over q-values by averaging
        # over the iterations of the training.
        # Here we have the q and action probs for all actions
        # so we calculate the state-value as the expectation of the q-values
        v_1 = batch_matmul(probs_a_t, (q_t - alpha * logprobs_a_t))
        q_target = jax.lax.stop_gradient(
            r_t + terminal_tm1 * self.hparams.discount * v_1
        )
        critic_loss = optax.l2_loss(qA_tm1, q_target) + optax.l2_loss(qB_tm1, q_target)
        critic_loss = jnp.asarray(critic_loss)

        # temperature target:$ H(\pi(s_t)) = 0.98 * -\sum_a \pi(a|s_t) \log(\pi(a|s_t))$
        target_entropy = 0.98 * (-jnp.log(1 / self.hparams.action_space.ndim))
        logprobs_a_tm1 = jax.lax.stop_gradient(logprobs_a_tm1)
        probs_a_tm1 = jax.lax.stop_gradient(probs_a_tm1)
        temperature_loss = -(temperature * logprobs_a_tm1 + target_entropy)
        temperature_loss = batch_matmul(probs_a_tm1, temperature_loss)

        loss = actor_loss + critic_loss + temperature_loss
        aux = (actor_loss, critic_loss, temperature_loss, policy_entropy, alpha)
        return loss, aux



# Copyright 2023 The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax
import jax.numpy as jnp
from jax.core import Shape

from .image import greyscale, imresize


def preprocess_atari(x):
    """Preprocessing function from
    Mnih, V., 2015, https://www.nature.com/articles/nature14236
    """
    # depthwise max pooling to remove flickering
    x = jax.lax.reduce_window(
        x, -jnp.inf, jax.lax.max, (2, 1, 1, 1), (1, 1, 1, 1), "SAME"
    )
    return greyscale(imresize(x, (84, 84)))


def preprocess_minigrid(x, size: Shape = (56, 56)):
    """Refer to the minigrid implementation at:
    https://github.com/Farama-Foundation/Minigrid
    """
    return imresize(x / 255, size=size, channel_first=False)



from typing import Tuple
import flax.linen as nn
from flax.core.scope import VariableDict as Params
from jax import Array
import jax.numpy as jnp
import optax
import rlax

from .mdp import TERMINATION, Timestep


def flatten_timesteps(timesteps: Timestep, discount: float = 1.0) -> Tuple[Array, Array, Array, Array, Array, Array]:
    s_tm1 = timesteps.observation[:-1]
    s_t = timesteps.observation[1:]
    a_tm1 = timesteps.action[:-1][0]  # [0] because scalar
    r_t = timesteps.reward[:-1][0]  # [0] because scalar
    terminal_tm1 = timesteps.step_type[:-1] != TERMINATION
    discount_t = discount ** timesteps.t[:-1][0]  # [0] because scalar
    return s_tm1, s_t, a_tm1, r_t, terminal_tm1, discount_t

def dqn_loss(
    timesteps: Timestep,
    critic: nn.Module,
    params: Params,
    params_target: Params,
    discount: float = 0.99,
) -> Array:
    s_tm1, s_t, a_tm1, r_t, terminal_tm1, discount_t = flatten_timesteps(timesteps, discount)

    q_tm1 = jnp.asarray(critic.apply(params, s_tm1))
    q_t = jnp.asarray(critic.apply(params_target, s_t)) * terminal_tm1

    td_error = rlax.q_learning(
        q_tm1, a_tm1, r_t, discount_t, q_t, stop_target_gradients=True
    )
    td_error = jnp.asarray(td_error)
    td_loss = jnp.mean(0.5 * td_error**2)
    return td_loss


def double_dqn_loss(
    timesteps: Timestep,
    critic: nn.Module,
    params: Params,
    params_target: Params,
    discount: float = 0.99,
):
    s_tm1, s_t, a_tm1, r_t, terminal_tm1, discount_t = flatten_timesteps(timesteps, discount)

    q_tm1 = jnp.asarray(critic.apply(params, s_tm1))
    a_t = jnp.argmax(jnp.asarray(critic.apply(params, s_t)) * terminal_tm1)
    q_t = critic.apply(params_target, s_t)
    q_target = r_t + discount_t * q_t[a_t]

    td_loss = optax.l2_loss(q_tm1[a_tm1] - q_target)
    return jnp.asarray(td_loss)


def soft_q_loss(
    timesteps: Timestep,
    critic: nn.Module,
    params: Params,
    params_target: Params,
    entropy: Array,
    discount: float = 0.99,
):
    s_tm1, s_t, a_tm1, r_t, terminal_tm1, discount_t = flatten_timesteps(timesteps, discount)


    raise NotImplementedError()
